/* Copyright (c) 2005 ACM/SIGDA
   Prepared by Jiang Hu, jhu@ee.tamu.edu
*/

#include "riata.h"
#include <string>

///////////////////////////////////////////////////
// Routines for Node 
//////////////////////////////////////////////////
void Node::dump() const
{
  if ( type == SOURCE )
    cout << "  source ";
  else if ( type == SINK )
    cout << "  sink ";
  else if ( type == STEINER )
    cout << "  steiner ";
  else
    cout << "  Unknown node type ";

  cout << id << " ( " << x << " " << y << " ) ";
  if ( type == SINK )
    cout << "cap " << cap << " rat " << rat << " "; 
  if ( parent != NULL )
    cout << "parent " << parent->ID() << " ";
  if ( leftChild != NULL )
    cout << "leftChild " << leftChild->ID() << " ";
  if ( rightChild != NULL )
    cout << "rightChild " << rightChild->ID() << " ";
  cout << endl;
}

///////////////////////////////////////////////////
// Routines for BufSolution
//////////////////////////////////////////////////
// Propagate the solution from its current node
// to its parent node
void BufSolution::addWire( const Node& parent )
{
  long edgeLen = root.dist( parent ); 
  long edgeRes = edgeLen * RIATA::wireRes;
  long edgeCap = edgeLen * RIATA::wireCap;
  long loadCap = edgeCap + root.Cap();
  long rat = root.RAT() - edgeRes * ( root.Cap() + edgeCap/2 );  
  root = parent;
  root.setCap( loadCap );
  root.setRAT( rat ); 
}

// Add buffer at current node 
void BufSolution::addBuf( long child )
{
  long rat = root.RAT() - RIATA::bufRes * root.Cap();
  root.setCap( RIATA::bufCap );
  root.setRAT( rat ); 
  bufNodes.push_back( pair<long, long>(root.ID(), child) );
}

void BufSolution::addDriver() 
{
  long rat = root.RAT() - RIATA::driverRes * root.Cap();
  root.setRAT( rat ); 
  root.setCap( RIATA::bufCap );
}

// The solution is replaced by the merged solution
void BufSolution::merge( BufSolution& s )
{
  long rat = min( root.RAT(), s.RAT() );
  long cap = root.Cap() + s.Cap();
  root.setCap( cap );
  root.setRAT( rat );
 
  list< pair<long, long> > tmpBuf( s.BufNodes() );   
  bufNodes.splice( bufNodes.end(), tmpBuf );
  list< pair<long, Point> > tmpSteiner( s.SteinerLocations() ); 
  steinerLocations.splice( steinerLocations.end(), 
		           tmpSteiner );  
}

void BufSolution::recordSteinerChange()
{
  Point p( root.X(), root.Y() );
  steinerLocations.push_back( pair<long, Point>( root.ID(), p ) ); 
}

void BufSolution::dump() const
{
  cout << "BufSolution:\n";
  cout << "  RAT " << root.RAT() << "   cap " << root.Cap(); 
  cout << "   root " << root.ID() << endl;

  cout << "  List of buffered nodes:\n";
  list< pair<long, long> >::const_iterator i;
  for ( i = bufNodes.begin(); i != bufNodes.end(); i++ ) {
    cout << "    " << (*i).first << " drives child ";
    cout << (*i).second << endl; 
  }

  cout << "  List of Steiner location changes:\n"; 
  list< pair<long, Point> >::const_iterator j;
  for ( j = steinerLocations.begin(); j != steinerLocations.end(); j++ ) {
    cout << "    " << (*j).first;
    cout << " ( " << (*j).second.X() << " " 
         << (*j).second.Y() << " )\n";
  }
}


///////////////////////////////////////////////////
// Routines for BufSolutionSet
//////////////////////////////////////////////////

void BufSolutionSet::propagate( const Node& parent ) 
{
  assert( !( solutionList.empty() ) );
  list<BufSolution>::iterator i;
  for ( i = solutionList.begin(); i != solutionList.end(); i++ ) {
    (*i).addWire( parent ); 
  }
  prune();
}

// Among existing solutions, find the one with max RAT
// after buffer insertion. Add this buffered solution
// to the set.  
void BufSolutionSet::addBuf( long child )
{ 
  list<BufSolution>::iterator i;
  i = solutionList.begin();
  BufSolution* bestCandidate = &(*i);
  long maxRAT = (*i).RAT() - RIATA::bufRes*(*i).Cap();
  while ( ++i != solutionList.end() ) {
    long rat = (*i).RAT() - RIATA::bufRes*(*i).Cap();
    if ( rat > maxRAT ) {
      maxRAT = rat;
      bestCandidate = &(*i);
    }
  }
 
  solutionList.push_back( *bestCandidate ); 
  solutionList.back().addBuf( child );
  prune();
} 

void BufSolutionSet::addDriver() 
{ 
  list<BufSolution>::iterator i;
  for ( i = solutionList.begin(); i != solutionList.end(); i++ )
    (*i).addDriver();
}

// Replace the solution set with the merged solution set 
void BufSolutionSet::merge( BufSolutionSet& rightSolutions )
{
#if DEBUG
  cout << "Will merge following two sets of solutions\n";
  this->dump();
  rightSolutions.dump();
#endif 
    
  list<BufSolution> leftList( solutionList );
  solutionList.erase( solutionList.begin(), solutionList.end() );
  list<BufSolution>& rightList = rightSolutions.SolutionList();

  while ( !leftList.empty() && !rightList.empty() ) {
    BufSolution& leftTop = leftList.front();
    BufSolution& rightTop = rightList.front();
    if ( leftTop.RAT() < rightTop.RAT() ) {  
      leftTop.merge( rightTop );
      solutionList.push_back( leftTop );
      leftList.pop_front();
    } else {
      rightTop.merge( leftTop );
      solutionList.push_back( rightTop );
      rightList.pop_front();
    }
  }
}

// After pruning, the solution with the max RAT should
// be at the end.
const BufSolution& BufSolutionSet::bestFinalSolution() 
{
  prune();
  return solutionList.back();
}

void BufSolutionSet::recordSteinerChange()
{
  list<BufSolution>::iterator i;
  for ( i = solutionList.begin(); i != solutionList.end(); i++ ) {
    (*i).recordSteinerChange();
  }
}

void BufSolutionSet::prune()
{
  solutionList.sort( lessCapSolution() );
  list<BufSolution>::iterator i, j, toBeErased;
  i = j = solutionList.begin();
  j++;
  while ( j != solutionList.end() ) {
    if ( (*i).RAT() > (*j).RAT() ) { // j is inferior
      toBeErased = j;
      j++;
      solutionList.erase( toBeErased );
    }
    else i = j++;
  }
}

void BufSolutionSet::dump() const
{
  cout << "BufSolutionSet\n";
  list<BufSolution>::const_iterator i;
  for ( i = solutionList.begin(); i != solutionList.end(); i++ ) {
    (*i).dump();
    cout << endl; 
  }
}

///////////////////////////////////////////////////
// Routines for RIATA 
//////////////////////////////////////////////////

long RIATA::driverRes( 1 );
long RIATA::wireRes( 1 );
long RIATA::wireCap( 1 );
long RIATA::bufRes( 1 );
long RIATA::bufCap( 1 );

// Read the tree into the node list, rectangles
// to the rect list, also technology parameters.  
RIATA::RIATA( const char* treeFile ) :
  root( NULL )
{
  string strBuf;
  ifstream inFile2;

  inFile2.open( treeFile, ios::in ); 
  while ( inFile2 >> strBuf ) {
    if ( strBuf == "source" ) {
      long i, x, y; 
      inFile2 >> i >> x >> y;
      nodes.push_back( Node( SOURCE, i, Point(x, y) ) );  
    }
    else if ( strBuf == "sink" ) {
      long i, x, y, c, q; 
      inFile2 >> i >> x >> y >> c >> q;
      nodes.push_back( Node( SINK, i, Point(x, y), c, q ) );  
    }
    else if ( strBuf == "steiner" ) {
      long i, x, y;
      inFile2 >> i >> x >> y; 
      nodes.push_back( Node( STEINER, i, Point(x, y) ) );  
    }
    else if ( strBuf == "edge" ) {
      long upEnd, downEnd;
      inFile2 >> upEnd >> downEnd;
      insertEdge( upEnd, downEnd );
    }  
    else if ( strBuf == "driverRes" ) {
      inFile2 >> driverRes;
    }
    else if ( strBuf == "wireRes" ) {
      inFile2 >> wireRes;
    }		
    else if ( strBuf == "wireCap" ) {
      inFile2 >> wireCap;
    }		
    else if ( strBuf == "bufRes" ) {
      inFile2 >> bufRes;
    }		
    else if ( strBuf == "bufCap" ) {
      inFile2 >> bufCap;
    }		
    else if ( strBuf == "rect" ) {
      long loX, loY, hiX, hiY;
      inFile2 >> loX >> loY;
      inFile2 >> hiX >> hiY;
      blockages.push_back( Rect( loX, loY, hiX, hiY ) );   
    }   
  }
  inFile2.close();

  NodeList::iterator i;
  for ( i = nodes.begin(); i != nodes.end(); i++ ) {
    if ( (*i).Type() == SOURCE ) {
      root = &(*i);
      break;
    }
  }
  assert( root != NULL ); 
}

// This is for inputing the Steiner tree. 
void RIATA::insertEdge( long upID, long downID )
{
  Node* up = NULL;
  Node* down = NULL;
  NodeList::iterator i;

  for ( i = nodes.begin(); i != nodes.end(); i++ ) {
    if ( (*i).ID() == upID )
      up = &(*i);
    else if ( (*i).ID() == downID )
      down = &(*i);
  }

  if ( up == NULL )  
    cout << "RIATA::insertEdge() - node " << upID << " not found\n"; 
  else if ( down == NULL )  
    cout << "RIATA::insertEdge() - node " << downID << " not found\n"; 
  else {
    down->setParent( up );
    if ( up->LeftChild() == NULL )
      up->setLeftChild( down );
    else
      up->setRightChild( down );
  }
}
 
void RIATA::doRIATA()
{
  BufSolutionSet solutions;
  findBufSolutions( *root, solutions ); 
  solutions.addDriver();
  bestSolution = solutions.bestFinalSolution(); 
}

// First check if the node 'n' overlaps any blockage.
// If so, find an alternative location for it. 
pair<bool, Point> RIATA::blockedAndAlternative( const Node& n ) const
{
  long newX(0), newY(0); 
  RectList::const_iterator i;
  bool isBlocked = false;

  for ( i = blockages.begin(); i != blockages.end(); i++ ) {
    if ( (*i).containPoint( n ) ) {
      isBlocked = true;
      newX = labs( n.X() - (*i).loX() ) <  labs( n.X() - (*i).hiX() ) ?
             (*i).loX() : (*i).hiX(); 
      newY = labs( n.Y() - (*i).loY() ) <  labs( n.Y() - (*i).hiY() ) ?
             (*i).loY() : (*i).hiY(); 
      if ( labs( n.X() - newX ) < labs( n.Y() - newY ) )
	newY = n.Y();
      else
	newX = n.X();	
      break;
    }
  }

  return pair<bool, Point>( isBlocked, Point(newX, newY) ); 
}

void RIATA::dumpNodes() const
{
  cout << "\nNode list: \n";
  NodeList::const_iterator i;

  for ( i = nodes.begin(); i != nodes.end(); i++ ) {
    (*i).dump();
  }
}

void RIATA::dumpBlockages() const
{
  cout << "\nBlockage list: \n";
  RectList::const_iterator i;

  for ( i = blockages.begin(); i != blockages.end(); i++ ) {
    (*i).dump();
  }
}

void RIATA::dumpParameters() const
{
  cout << "\nDriver resistance: " << driverRes << endl;
  cout << "Wire resistance: " << wireRes << endl; 
  cout << "Wire capacitance: " << wireCap << endl; 
  cout << "Buffer resistance: " << bufRes << endl;
  cout << "Buffer capacitance: " << bufCap << "\n\n"; 
}

// This function writes the result to stdout. DO NOT CHANGE IT.
void RIATA::output() 
{
  cout.setf( ios::fixed, ios::floatfield );
  cout.precision( 1 );

  cout << "Slack  " << bestSolution.RAT() << endl; 

  // Buffer node lost
  const list< pair<long, long> >& L = bestSolution.BufNodes();
  list< pair<long, long> >::const_iterator i;
  for ( i = L.begin(); i != L.end(); i++ ) {
    cout << "BUF " << (*i).first << " drives child ";
    cout << (*i).second << endl; 
  }
 
  // List of Steiner location changes
  const list< pair<long, Point> >& S = bestSolution.SteinerLocations();
  list< pair<long, Point> >::const_iterator j;
  for ( j = S.begin(); j != S.end(); j++ ) {
    cout << "STEINER_LOC_CHANGE " << (*j).first;
    cout << " ( " << (*j).second.X() << " " << (*j).second.Y() << " )\n";
  }
}
