/* Copyright (c) 2005 ACM/SIGDA
   Prepared by Jiang Hu, jhu@ee.tamu.edu
*/

#ifndef RIATA_H
#define RIATA_H

#include <iostream>
#include <fstream>
#include <string>
#include <assert.h>
#include <math.h>
#include <vector>
#include <list>
#include <map>

using namespace std;

class Point {
  public:

    // Constructors
    Point() : x( 0 ), y( 0 ) {}
    Point( long xi, long yi ) : x( xi ), y( yi ) {} 

    // Accessors 
    inline long X() const { return x; }
    inline long Y() const { return y; }

    // Mutators 
    inline void setLocation( const Point& p ) {
      x = p.X(); y = p.Y(); }
    inline void setX( long xi ) { x = xi; }
    inline void setY( long yi ) { y = yi; }   

    // Others
    inline bool operator== ( const Point& rhs ) const {
      return x == rhs.x && y == rhs.y; }
    inline long dist ( const Point& p ) const {
      return labs( x - p.X() ) + labs( y - p.Y() ); }  

  protected:
    long x, y;
};

class Rect {
  public:
    // Constructors 
    Rect( long loX, long loY, long hiX, long hiY ) : 
      lo( loX, loY ), hi( hiX, hiY ) {}  

    // Accessors
    inline long loX() const { return lo.X(); }
    inline long loY() const { return lo.Y(); }
    inline long hiX() const { return hi.X(); }
    inline long hiY() const { return hi.Y(); }

    inline bool containPoint( const Point& p ) const {
      return lo.X() < p.X() && p.X() < hi.X() && 
             lo.Y() < p.Y() && p.Y() < hi.Y(); } 

    inline void dump() const {
      cout << "  rect( " << lo.X() << ", " << lo.Y()
           << " ) - ( " << hi.X() << ", " << hi.Y() << " )\n"; } 
  private:
    Point lo, hi;
}; 

enum NodeType {
    SOURCE,
    SINK,
    STEINER 
};

class Node : public Point {
  public:
    // Constructors
    Node() : Point( 0, 0 ), type( STEINER ), id( 0 ), cap( 0 ), rat( 0 ), 
      parent( NULL ), leftChild( NULL ), rightChild( NULL ) {} 
    Node( NodeType tp, long i, Point( p ),  long c = 0, long q = 0 ): 
      Point( p ), type( tp ), id( i ), cap( c ), rat( q ), 
      parent( NULL ), leftChild( NULL ), rightChild( NULL ) {} 

    // Accessors 
    inline NodeType Type() const { return type; }
    inline long ID()  const { return id; }
    inline long Cap() const { return cap; } 
    inline long RAT() const { return rat; } 
    inline Node* Parent()     const { return parent; }
    inline Node* LeftChild()  const { return leftChild; }
    inline Node* RightChild() const { return rightChild; }
    void dump() const; 

    // Settings
    inline void setCap( long c ) { cap = c; }
    inline void setRAT( long q ) { rat = q; }
    inline void setParent( Node* const p )     { parent = p; } 
    inline void setLeftChild( Node* const n )  { leftChild = n; }
    inline void setRightChild( Node* const n ) { rightChild = n; }

  private:
    NodeType type;
  
    long id;
    long cap;   
    long rat; 
    
    Node* parent;  
    Node* leftChild;
    Node* rightChild;
};

class BufSolution {
  public:
    // Constructors
    BufSolution() {}  
    BufSolution( const Node& n ): root( n ) {} 

    // Accessors 
    inline long Cap() const { return root.Cap(); } 
    inline long RAT() const { return root.RAT(); } 
    inline list< pair<long, long> >& BufNodes() { return bufNodes; }
    inline list< pair<long, Point> >& SteinerLocations() { 
      return steinerLocations; }

    // Settings
    inline void setCap( long c ) { root.setCap( c ); }
    inline void setRAT( long q ) { root.setRAT( q ); }

    // Others
    void addWire( const Node& );  
    void addBuf( long ); 
    void addDriver();
    void merge( BufSolution& );
    void recordSteinerChange();    
    void dump() const;
    inline bool operator< ( const BufSolution& s ) const {
      return Cap() < s.Cap(); }

  private:
    Node root; 

    list< pair<long, long> > bufNodes; // id of buf node and its child 
                                         
    list< pair<long, Point> > steinerLocations;
}; 

class BufSolutionSet {
  public:
    inline list<BufSolution>& SolutionList() { return solutionList; }
    inline void addSolution( BufSolution s ) { solutionList.push_back( s ); }
    void propagate( const Node& );
    void addBuf( long );
    void addDriver();
    void merge( BufSolutionSet& );
    const BufSolution& bestFinalSolution(); 
    void recordSteinerChange();
    inline void combine( BufSolutionSet& B ) { 
      solutionList.merge( B.SolutionList() ); }
    void dump() const;

  private:
    list<BufSolution> solutionList; 

    void prune();
};

class RIATA {
  public:
    RIATA( const char* );
    void doRIATA();

    void dumpNodes() const;
    void dumpBlockages() const;
    void dumpParameters() const;
    inline void dumpBestSolution() const {
      bestSolution.dump(); }
    void output();

    typedef vector<Node> NodeList;
    typedef vector<Rect> RectList; 

    friend class BufSolution;
    friend class BufSolutionSet;

  private:
    static long driverRes;
    static long bufRes;
    static long bufCap;
    static long wireRes;
    static long wireCap;

    Node* root;
    NodeList nodes;
    RectList blockages;
    BufSolution bestSolution;

    pair<bool, Point> blockedAndAlternative( const Node& n ) const;
    void findBufSolutions( const Node&, BufSolutionSet& );
    void insertEdge( long, long );
};

class lessCapSolution {
  public:
    bool operator() ( BufSolution& A, BufSolution& B ) 
    {
      return A.Cap() < B.Cap();
    }
};

#endif
