/* Copyright (c) 2005 ACM/SIGDA
   Prepared by Geert Janssen, geert@us.ibm.com
*/

#ifndef BDD_API_H
#define BDD_API_H

/* Need this from CUDD package: */
#include <CUDD/util.h>
#include <CUDD/cudd.h>

#if defined __cplusplus
extern "C" {
#endif

/* -------------------------------------------------------------------------*/
/*			    BDD Package Interface			    */
/* -------------------------------------------------------------------------*/

/* NOTE: unlike CUDD, this interface ensures that all functions that return
   a BDD will indeed return a fresh reference. This implies that it is the
   user's responsibility to free up that reference again when the BDD is
   no longer needed. Not ever calling Bdd_free() is OK but might cause
   a lot of unnecessary memory allocation. You are warned!

   The main data type is "BDD", the manager's type is "BDD_Mgr".
   All functions start with the prefix "Bdd_".
*/

/* Type for the manager object: */
typedef DdManager* BDD_Mgr;
/* Type for a BDD object: */
typedef DdNode* BDD;

/* Initializes BDD package and returns manager for it. */
BDD_Mgr Bdd_init(void);

/* Quits invocation of BDD package via manager. */
void Bdd_quit(BDD_Mgr m);

/* Duplicates BDD f (adds another reference). */
BDD Bdd_dup(BDD_Mgr m, BDD f);

/* Tests whether f is the 0 BDD. */
int Bdd_0_p(BDD_Mgr m, BDD f);

/* Returns the 0 BDD. */
BDD Bdd_0(BDD_Mgr m);

/* Tests whether f is the 1 BDD. */
int Bdd_1_p(BDD_Mgr m, BDD f);

/* Returns the 1 BDD. */
BDD Bdd_1(BDD_Mgr m);

/* Returns the BDD for the projection on variable v. */
BDD Bdd_var(BDD_Mgr m, int v);

/* Returns fresh ref for complement BDD of f. */
BDD Bdd_not(BDD_Mgr m, BDD f);

/* Returns fresh ref for conjunction of BDDs f and g. */
BDD Bdd_and(BDD_Mgr m, BDD f, BDD g);

/* Returns conjunction over all BDDs in array A of length n. */
BDD Bdd_and_vec(BDD_Mgr m, BDD A[], int n);

/* Returns fresh ref for disjunction of BDDs f and g. */
BDD Bdd_or(BDD_Mgr m, BDD f, BDD g);

/* Returns disjunction over all BDDs in array A of length n. */
BDD Bdd_or_vec(BDD_Mgr m, BDD A[], int n);

/* Returns fresh ref for XNOR of BDDs f and g. */
BDD Bdd_equiv(BDD_Mgr m, BDD f, BDD g);

/* Frees up a reference to BDD f. */
void Bdd_free(BDD_Mgr m, BDD f);

/* Frees all BDDs in array A of length n. */
void Bdd_free_vec(BDD_Mgr m, BDD A[], int n);

/* Returns count of minterms in BDD f assuming space of n variables. */
int Bdd_nr_minterms(BDD_Mgr m, BDD f, int n);

/* Swaps all x[i] for y[i] variables (both ways) in BDD f.
   x and y are arrays of variable BDDs of length n.
*/
BDD Bdd_swap_vars(BDD_Mgr m, BDD f, BDD *x, BDD *y, int n);

/* Returns existential quantification in BDD f of all variables in vs cube. */
BDD Bdd_exist(BDD_Mgr m, BDD f, BDD vs);

/* Returns universal quantification in BDD f of all variables in vs cube. */
BDD Bdd_univ(BDD_Mgr m, BDD f, BDD vs);

#ifdef __cplusplus
}
#endif

#endif /* BDD_API_H */
