/* Copyright (c) 2005 ACM/SIGDA
   Prepared by Geert Janssen, geert@us.ibm.com
*/

#ifndef TREE_H
#define TREE_H

/* ------------------------------------------------------------------------ */
/* INCLUDES								    */
/* ------------------------------------------------------------------------ */

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>

#if defined __cplusplus
extern "C" {
#endif

/* ------------------------------------------------------------------------ */
/* DEFINES								    */
/* ------------------------------------------------------------------------ */

/* Tree node field access macros: */
#define T_TYPE(t)		((t)->type)
#define T_LEFT(t)		((t)->left)
#define T_RIGHT(t)		((t)->right)

/* Tree constructor macros: */
#define tree_mk_const(c)	tree_mk((c),NULL,NULL)
#define tree_mk_0()		tree_mk_const('0')
#define tree_mk_1()		tree_mk_const('1')
#define tree_mk_var(v)		tree_mk((v),NULL,NULL)
#define tree_mk_not(t)		tree_mk('!',(t),NULL)
#define tree_mk_and(t1,t2)	tree_mk('&',(t1),(t2))
#define tree_mk_or(t1,t2)	tree_mk('+',(t1),(t2))

#define tree_mk_seq(t1,t2)	tree_mk(';',(t1),(t2))
#define tree_mk_nsf(v,t)	tree_mk('@',tree_mk_var(v),(t))
#define tree_mk_inv(t)		tree_mk('?',(t),NULL)

/* Generic predicate for tree node types: */
#define tree_p(t, x)		(T_TYPE(t) == (x))

/* Predicates for operators and various other tree node types: */
#define tree_0_p(t)		tree_p(t, '0')
#define tree_1_p(t)		tree_p(t, '1')
#define tree_const_p(t)		(tree_p(t, '0') || tree_p(t, '1'))
#define tree_var_p(t)		isalpha(T_TYPE(t))
#define tree_not_p(t)		tree_p(t, '!')
#define tree_and_p(t)		tree_p(t, '&')
#define tree_or_p(t)		tree_p(t, '+')

#define tree_seq_p(t)		tree_p(t, ';')
#define tree_nsf_p(t)		tree_p(t, '@')

/* ------------------------------------------------------------------------ */
/* TYPE DEFINITIONS							    */
/* ------------------------------------------------------------------------ */

/* Data structure for finite state transition system:
   A binary tree with internal nodes that represent next-state functions
   and expressions, and leaf nodes that represent constants and variables.
   The type of a node is encoded by a character:

   For internal (operator) nodes:
   '!'	  : a  NOT node, operand is left branch; right branch is NULL;
   '&'	  : an AND node, operands are left and right branches;
   '+'	  : an	OR node, operands are left and right branches.

   ';'	  : a sequence node, element is left branch, rest via right branch.
   '@'	  : a next-state function node, var on left, expr on right.

   For leaf nodes (left and right are both NULL):
   '0'	  : a constant 0 (false) node;
   '1'	  : a constant 1 (true) node;
   letter : (case-sensitive) a variable node.

   Any other character for type is impossible, and this need not be tested.
*/
typedef struct node *Tree;
struct node {
  int type;
  Tree left;
  Tree right;
};

/* ------------------------------------------------------------------------ */
/* FUNCTION PROTOTYPES							    */
/* ------------------------------------------------------------------------ */

/* Returns the tree constructed by creating a fresh root node of type
   `type' and giving it the children `left' and `right'.
   (You might prefer to use the specialized constructor macros instead.)
*/
Tree tree_mk(int type, Tree left, Tree right);

/* Prints tree `t' to stdout in infix notation. */
void tree_print(Tree t);

/* Frees the tree `t'. No action for NULL argument. */
void tree_free(Tree t);

#ifdef __cplusplus
}
#endif

#endif /* TREE_H */
