/* Copyright (c) 2005 ACM/SIGDA
   Prepared by Kedarnath Balakrishnan, bala@nec-labs.com
*/

/* This is ckt.c */
#include <stdio.h>
#include <stdlib.h>
#include "define.h"

/*------------------------------------------------------------------------*/
NSTRUC *Node;			/* dynamic array of nodes */
NSTRUC **Pinput;		/* pointer to array of primary inputs */
NSTRUC **Poutput;		/* pointer to array of primary outputs */
NSTRUC **seq;			/* signal propagation order */
unsigned Nnodes;		/* number of nodes */
unsigned Npi;			/* number of primary inputs */
unsigned Npo;			/* number of primary outputs */
unsigned Ngate;			/* number of gates */
int Nints;			/* number of integers for fault list */
int Nflt;			/* number of faults in the fault file */
int *Fl;			/* temprary fault list used in fsim */
fault_list *fault_set;		/* target fault list */

/* global variables and functions */

int comment(char buf[]);
int allocate(void);
unsigned levelize(void);
unsigned level(NSTRUC *);
unsigned lev(NSTRUC *);

/*-----------------------------------------------------------------------
input: circuit description file name
output: nothing
called by: main
description:
This routine reads in the circuit description file and set up all the
required data structure. It first checks if the file exists, then it
sets up a mapping table, determines the number of nodes, PI's and PO's,
allocates dynamic data arrays, and fills in the structural information
of the circuit. In the ISCAS circuit description format, only upstream
nodes are specified. Downstream nodes are implied. However, to facilitate
forward implication, they are also built up in the data structure.
To have the maximal flexibility, three passes through the circuit file
are required: the first pass to determine the size of the mapping table
, the second to fill in the mapping table, and the third to actually
set up the circuit information. These procedures may be simplified in
the future.
-----------------------------------------------------------------------*/
int
cread(char *cp)
{
  char buf[MAXLINE];
  int ntbl, nd, tp, pi = 0, line, po = 0, ty;
  unsigned int i, fi, fo;
  int *tbl;
  FILE *fd;
  NSTRUC *np;

  sscanf(cp, "%s", buf);
  if ((fd = fopen(buf, "r")) == NULL) {
    printf("File %s does not exist!\n", buf);
    exit(-1);
  }
  Nnodes = Npi = Npo = ntbl = 0;
  while (fgets(buf, MAXLINE, fd) != NULL) {
    if (sscanf(buf, "%d %d", &tp, &nd) == 2) {
      if (ntbl < nd)
	ntbl = nd;
      Nnodes++;
      if (tp == PI)
	Npi++;
      else if (tp == PO)
	Npo++;
    }
  }
  tbl = (int *) malloc(++ntbl * sizeof(int));

  fseek(fd, 0L, 0);
  i = 0;
  while (fgets(buf, MAXLINE, fd) != NULL) {
    if (sscanf(buf, "%d %d", &tp, &nd) == 2)
      tbl[nd] = i++;
  }
  allocate();

  fseek(fd, 0L, 0);
  line = 0;
  while (fscanf(fd, "%u %d", &i, &nd) != EOF) {
    line++;
    np = &Node[tbl[nd]];
    np->num = nd;
    switch (i) {
    case PO:
      Poutput[po++] = np;

    case GATE:
      fscanf(fd, "%d %u %u", &ty, &np->fout, &np->fin);
      np->type = ty;
      np->unodes = (NSTRUC **) malloc(np->fin * sizeof(NSTRUC *));
      np->dnodes = (NSTRUC **) malloc(np->fout * sizeof(NSTRUC *));
      for (fi = 0; fi < np->fin; fi++) {
	fscanf(fd, "%d", &nd);
	np->unodes[fi] = &Node[tbl[nd]];
      }
      for (fo = 0; fo < np->fout; fo++)
	np->dnodes[fo] = NULL;
      break;

    case PI:
      Pinput[pi++] = np;
      fscanf(fd, "%d %u %u", &ty, &np->fout, &np->fin);
      np->type = ty;
      np->dnodes = (NSTRUC **) malloc(np->fout * sizeof(NSTRUC *));
      for (fo = 0; fo < np->fout; fo++)
	np->dnodes[fo] = NULL;
      break;

    case FB:
      np->fout = np->fin = 1;
      fscanf(fd, "%d %d", &ty, &nd);
      np->type = ty;
      np->unodes = (NSTRUC **) malloc(sizeof(NSTRUC *));
      np->dnodes = (NSTRUC **) malloc(sizeof(NSTRUC *));
      np->unodes[0] = &Node[tbl[nd]];
      np->dnodes[0] = NULL;
      break;

    default:
      printf("Unknown node type: %u\n", line);
      break;
    }
  }
  for (i = 0; i < Nnodes; i++) {
    for (fi = 0; fi < Node[i].fin; fi++) {
      np = Node[i].unodes[fi];
      fo = 0;
      while ((np->dnodes)[fo] != NULL)
	fo++;
      np->dnodes[fo] = &Node[i];
    }
  }
  free(tbl);
  fclose(fd);
  return (ntbl);
}

/*------------------------------------------------------------------------
input : a pointer to fault description file
output: nothing
called by: main
description: read fault description file and make fault list
-------------------------------------------------------------------------*/

int
read_fault(FILE *ff)
{
  fault_list *new;
  unsigned i, j;
  char c, buf[MAXLINE];
  Nflt = 0;

  while (fgets(buf, MAXLINE, ff) != NULL) {
    if (sscanf(buf, "%u %c", &i, &c) != 2) {
      printf("Not enough fault description specified!\n");
      exit(-1);
    }

    j = 0;
    while (j < Nnodes && i != Node[j].num)
      j++;
    if (j >= Nnodes) {
      printf("Line %d does not exist.\n", i);
      exit(-1);
    } else if (!Isbool(c)) {
      printf("Illegal Boolean value \"%c\"!\n", c);
      exit(-1);
    }

    new = (fault_list *) malloc(sizeof(fault_list));
    new->node = &Node[j];
    new->sa01 = c;
    new->next = fault_set;
    new->prev = NULL;
    if (fault_set != NULL)
      fault_set->prev = new;
    fault_set = new;
    Nflt++;
  }
  return 0;
}

/*-----------------------------------------------------------------------
input: nothing
output: nothing
called by: main
description: checks if the current line is a comment line
-----------------------------------------------------------------------*/

int
comment(char buf[])
{
  if (buf[0] == '|')
    return (1);
  else
    return (0);
}

/*-----------------------------------------------------------------------
input: nothing
output: nothing
called by: cread
description:
This routine allocatess the memory space required by the circuit
description data structure. It allocates the dynamic arrays Node,
Node.flist, Node, Pinput, Poutput, and Tap. It also set the default
tap selection and the fanin and fanout to 0.
-----------------------------------------------------------------------*/
int
allocate(void)
{
  unsigned int i;

  Nints = Nnodes / INTSIZE + ((Nnodes % INTSIZE) ? 1 : 0);
  Node = (NSTRUC *) malloc(Nnodes * sizeof(NSTRUC));
  Pinput = (NSTRUC **) malloc(Npi * sizeof(NSTRUC *));
  Poutput = (NSTRUC **) malloc(Npo * sizeof(NSTRUC *));
  Fl = (int *) malloc(Nints * sizeof(int));
  for (i = 0; i < Nnodes; i++) {
    Node[i].indx = i;
    Node[i].flist = (int *) malloc(Nints * sizeof(int));
    Node[i].fin = Node[i].fout = 0;
  }
  return 0;
}

/*-----------------------------------------------------------------------
input: nothing
output: maximal level of the circuit
called by: main
description:
This routine calculates the levels of all the nodes in the circuit.
The level of PI's is 0.
-----------------------------------------------------------------------*/
unsigned
levelize(void)
{
  unsigned int i;
  int rlev, maxlevel = 0;
  NSTRUC *np;

  for (i = 0; i < Nnodes; i++) {
    np = &Node[i];
    if (np->type == IPT)
      np->level = 0;
    else
      np->level = -1;
  }
  for (i = 0; i < Npo; i++) {
    rlev = lev(Poutput[i]);
    if (rlev > maxlevel)
      maxlevel = rlev;
  }
  return (maxlevel);
}

/*-----------------------------------------------------------------------
input: pointer to a node in the array Node
output: level of the gate output
called by: levelize
description:
This recursive routine calculates the level of the given node. The
level of a gate output is the maximal level of the gate inputs plus 1.
The level of stem branches is the same as the one of the stem.
-----------------------------------------------------------------------*/
unsigned
lev(NSTRUC *np)
{
  unsigned int i;
  int rlev, maxlevel;

  if (np->level != -1)
    rlev = np->level;
  else {
    maxlevel = 0;
    for (i = 0; i < np->fin; i++) {
      rlev = lev(np->unodes[i]);
      if (rlev > maxlevel)
	maxlevel = rlev;
    }
    rlev = maxlevel + 1;
    np->level = rlev;
  }
  return (rlev);
}

/*-----------------------------------------------------------------------
input : maxlevel (maximum level of the circuit)
output:  nothing
called by: main
description:
find signal propagation sequence according to the level
in the circuit of each line
------------------------------------------------------------------------*/
int
find_seq(unsigned maxlevel)
{
  unsigned int i, j;
  int gate;

  gate = 0;
  for (i = 1; i <= maxlevel; i++)
    for (j = 0; j < Nnodes; j++) {
      if (!(Node[j].level))
	continue;
      else if (Node[j].level == (int) (i))
	seq[gate++] = &Node[j];
    }
  return 0;
}

/*-----------------------------------------------------------------------
input : none
output:  nothing
called by: main
description: prints circuit statistic information
------------------------------------------------------------------------*/

int
print_stat()
{
  unsigned int i;
  int gate, and, or, nand, nor, xor, not, brch, buff;

  gate = and = or = nand = nor = xor = not = brch = buff = 0;
  printf("Circuit Statistic Information\n");
  printf("================================================\n");
  printf("No. of Inputs: %d\n", Npi);
  printf("No. of outputs: %d\n", Npo);
  for (i = 0; i < Nnodes; i++) {
    switch (Node[i].type) {
    case BRCH:
      brch++;
      break;
    case XOR:
      xor++;
      break;
    case OR:
      or++;
      break;
    case NOR:
      nor++;
      break;
    case NOT:
      not++;
      break;
    case NAND:
      nand++;
      break;
    case AND:
      if (Node[i].fin == 1)
	buff++;
      else
	and++;
      break;
    default:
      printf("You should never reach this case in print_stat\n");
      exit(-1);
    }
  }
  gate = xor + or + nor + not + nand + and + buff;
  printf("No. of gates: %d\n", gate);
  printf("No. of XOR gates: %d\n", xor);
  printf("No. of OR gates: %d\n", or);
  printf("No. of NOR gates: %d\n", nor);
  printf("No. of NOT gates: %d\n", not);
  printf("No. of NAND gates: %d\n", nand);
  printf("No. of AND gates: %d\n", and);
  printf("No. of BUFFERs: %d\n", buff);
  printf("No. of BRANCHES: %d\n", brch);
  printf("No. of lines: %d\n", Nnodes);
  printf("================================================\n");
  return 0;
}

/*========================= End of program ============================*/
