/* Copyright (c) 2005 ACM/SIGDA
   Prepared by Kedarnath Balakrishnan, bala@nec-labs.com
*/

/* This is sim.c */
#include <stdlib.h>
#include "define.h"

/* global varibales */

VSTRUC Tgfault;			/* target fault */
NSTRUC *Node;			/* dynamic array of nodes */
NSTRUC **Pinput;		/* pointer to array of primary inputs */
NSTRUC **Poutput;		/* pointer to array of primary outputs */
NSTRUC **seq;			/* seqnal propagation order */
node_list *Dfront;		/* D-frontier */
unsigned Nnodes;		/* number of nodes */
unsigned Npi;			/* number of primary inputs */
unsigned Npo;			/* number of primary outputs */
unsigned Ngate;			/* number of gates */
int Nints;                      /* number of integers for fault list */
int Nflt;			/* number of faults in the fault file */
int *Fl;			/* temprary fault list used in fsim */
int *det_sa1_fault;		/* s-a-1 fault set detected */
int *det_sa0_fault;		/* s-a-0 fault set detected */
fault_list *fault_set;		/* target fault list */

/* functions */
int  eval(void);
char simply(NSTRUC *np);
int  fdrive(void);
fault_list *search_fault(NSTRUC *np, char val);
char inv(char);
char xor_val(unsigned vc[]);

/*-----------------------------------------------------------------------
input: an input vector and output stream
output: nothing
called by: main
description:
The routine perform true value simulation for current input vector. It
resets all lines to 'X', setups input vectors, and calls recursive
"eval" routine to calculates every line value. Finally, it prints out
the result in the output stream.
-----------------------------------------------------------------------*/
int
tsim(char *vec)
{
  unsigned int i;
  int j;
  NSTRUC *np;

  j = 0;
  for (i = 0; i < Nnodes; i++) {
    np = &Node[i];
    if (np->type == IPT)
      np->val = vec[j++];
    else
      np->val = 'X';
  }
  eval();
  return 0;
}

/*-----------------------------------------------------------------------
input: pointer to a node in the array Node
output: implied value of the gate output
called by: tsim
description:
This recursive routine calculates the implied value for the node by
calling "simply".
-----------------------------------------------------------------------*/
int
eval(void)
{
  unsigned i, fout;
  char rval;
  NSTRUC *np;

  for (i = 0; i < Npi; i++) {
    np = Pinput[i];
    rval = np->val;
  }

  for (i = 0; i < Ngate; i++) {
    np = seq[i];
    rval = simply(np);
    np->val = rval;
    if (np->fout == 0)
      fout = 1;
    else
      fout = np->fout;
  }
  return 0;
}

/*-----------------------------------------------------------------------
input: output stream
output: nothing
called by: main
description:
The routine performs deductive fault simulation for current assigned
input values. It zeros all fault lists, set the fault bit at PI's, and
then calls recursive "fdrive" routine for each Po's to propogates fault
lists. Finally, it outputs the number of faults detected by current input
vector.
-----------------------------------------------------------------------*/
unsigned
fsim(void)
{
  unsigned int i, j;
  NSTRUC *np;
  unsigned k = 0;
  fault_list *temp;
  int l;

  for (i = 0; i < Nnodes; i++) {
    np = &Node[i];
    Zeros(l, np->flist);
    if (np->type == IPT)
      Setbit(np->flist, i);
  }

  fdrive();
  for (i = 0; i < Npo; i++) {
    for (j = 0; j < Nnodes; j++)
      if (Node[j].val == '1') {	/* stuck-at-0 test */
	if (Bitset(Poutput[i]->flist, j)
	    && Bitset(det_sa0_fault, j)) {
	  np = &Node[j];
	  if ((temp = search_fault(np, '0')) != NULL)
	    k++;
	  Clearbit(det_sa0_fault, j);
	}			/* Wow, bad coding! Kartik. else ; */
      } else {			/* stuck-at-1 test */
	if (Bitset(Poutput[i]->flist, j)
	    && Bitset(det_sa1_fault, j)) {
	  np = &Node[j];
	  if ((temp = search_fault(np, '1')) != NULL)
	    k++;
	  Clearbit(det_sa1_fault, j);
	}			/* Wow, bad coding! Kartik. else ; */
      }
  }
  return (k);
}

/*-----------------------------------------------------------------------
input: pointer to a node in the array Node
output: nothing
called by: fsim
description:
This recursive routine performs deductive fault simulation on the given
node. It executes set operations on the fault list of the inputs and
produces a new fault list at the output. 
- -----------------------------------------------------------------------*/

int
fdrive(void)
{
  unsigned int k, i;
  int j, *fp;
  enum e_gtype tp;
  NSTRUC *np;

  for (k = 0; k < Ngate; k++) {
    np = seq[k];
    tp = np->type;
    fp = np->flist;
    /* if output value = 'X', then nullify the fault list of the output node */
    if (np->val == 'X' || np->val == 'x') {
      Zeros(j, fp);
    } else {
      if (tp == NOT || tp == BRCH)
	Funion(j, fp, fp, np->unodes[0]->flist);
      else {
	if (tp == OR || tp == NOR)
	  Ones(j, fp);
	else
	  Zeros(j, fp);
	for (i = 0; i < np->fin; i++) {
	  Fcopy(j, Fl, np->unodes[i]->flist);
	  if (tp == XOR)
	    Fxor(j, fp, fp, Fl);
	  else {
	    if (np->unodes[i]->val == '0')
	      Fcompliment(j, Fl, Fl);
	    if (tp == AND || tp == NAND)
	      Funion(j, fp, fp, Fl);
	    else
	      Fintersect(j, fp, fp, Fl);
	  }
	}
	if ((np->val == '0' && (tp == OR || tp == AND)) ||
	    (np->val == '1' && (tp == NAND || tp == NOR)))
	  Fcompliment(j, fp, fp);
      }
      Setbit(fp, np->indx);
    }
  }
  return 0;
}

/*-----------------------------------------------------------------------
input: pointer to a node in the array Node
output: implied value at the gate output
called by: imply, eval
description:
This routine performs the 5-valued logic simulation on the input gate.
It uses the concept of input counting for gate evaluation.
It also maintains the D-frontier used by podem and prints out the
D-frontier if it is updated. A gate is put in the D-frontier if the gate
output is 'X' but has one or more 'D' or 'B' on its inputs. Currently,
only NOT, AND, NAND, OR, NOR, and BRCH node types are implemented.
-----------------------------------------------------------------------*/
char
gimply(NSTRUC *np)
{
  unsigned int i, vc[5];
  int c, par, dchange;
  char cv, rval;
  node_list *dfp, *pdfp;

  switch (np->type) {
  case NOT:
    return (inv((np->unodes)[0]->val));
  case BRCH:
    return (np->unodes[0]->val);
  case AND:
    c = 0;
    par = 0;
    break;
  case NAND:
    c = 0;
    par = 1;
    break;
  case OR:
    c = 1;
    par = 0;
    break;
  case NOR:
    c = 1;
    par = 1;
    break;
  case XOR:
    c = 2;
    par = 0;
    break;
  default:
    exit(-1);
  }

  for (i = 0; i < 5; i++)
    vc[i] = 0;
  i = 0;
  while (i < np->fin && (!vc[c] || !(vc[3] && vc[4]))) {
    switch (np->unodes[i++]->val) {
    case '0':
      vc[0]++;
      break;
    case '1':
      vc[1]++;
      break;
    case 'X':
      vc[2]++;
      break;
    case 'D':
      vc[3]++;
      break;
    case 'B':
      vc[4]++;
      break;
    }
  }
  if (np->type == XOR) {
    if (vc[2])
      return ('X');
    else if (!vc[3] && !vc[4])
      return ((vc[1] % 2) ? '1' : '0');
    else
      return (xor_val(vc));
  }
  cv = c ? '1' : '0';
  if (vc[c] || (vc[3] && vc[4]))
    rval = cv;
  else if (vc[c ^ 1] == np->fin)
    rval = cv ^ 1;
  else if (vc[2] > 0)
    rval = 'X';
  else if (vc[3])
    rval = 'D';
  else if (vc[4])
    rval = 'B';

  dchange = 0;
  if (Dfront == NULL) {
    if ((vc[3] || vc[4]) && rval == 'X') {
      Dfront = (node_list *) malloc(sizeof(node_list));
      dfp = Dfront;
      dfp->ptr = np;
      dfp->next = NULL;
      dchange = 1;
    }				/* Bad coding. Kartik else ; */
  } else {
    pdfp = dfp = Dfront;
    while ((dfp != NULL) && (dfp->ptr != np)) {
      pdfp = dfp;
      dfp = dfp->next;
    }
    if (dfp == NULL) {
      if ((vc[3] || vc[4]) && rval == 'X') {
	pdfp->next = (node_list *) malloc(sizeof(node_list));
	dfp = pdfp->next;
	dfp->ptr = np;
	dfp->next = NULL;
	dchange = 1;
      }				/* Bad coding. Kartik else ; */
    } else if (rval != 'X' || (!(vc[3] || vc[4]) && rval == 'X')) {
      if (dfp == Dfront) {
	Dfront = pdfp->next;
	free(pdfp);
      } else {
	pdfp->next = dfp->next;
	free(dfp);
      }
      dchange = 1;
    }
  }

  return (par ? inv(rval) : rval);
}

/*-----------------------------------------------------------------------
input: pointer to a node in the array Node
output: implied value at the gate output
called by: imply, eval
description:
This routine performs the 5-valued logic simulation on the input gate.
It uses the concept of input counting for gate evaluation.
-----------------------------------------------------------------------*/
char
simply(NSTRUC *np)
{
  unsigned int i, vc[5];
  int c, par;
  char cv, rval;

  switch (np->type) {
  case NOT:
    return (inv((np->unodes)[0]->val));
  case BRCH:
    return (np->unodes[0]->val);
  case AND:
    c = 0;
    par = 0;
    break;
  case NAND:
    c = 0;
    par = 1;
    break;
  case OR:
    c = 1;
    par = 0;
    break;
  case NOR:
    c = 1;
    par = 1;
    break;
  case XOR:
    c = 2;
    par = 0;
    break;
  default:
    exit(-1);
  }

  for (i = 0; i < 5; i++)
    vc[i] = 0;
  i = 0;
  while (i < np->fin && (!vc[c] || !(vc[3] && vc[4]))) {
    switch (np->unodes[i++]->val) {
    case '0':
      vc[0]++;
      break;
    case '1':
      vc[1]++;
      break;
    case 'X':
      vc[2]++;
      break;
    case 'D':
      vc[3]++;
      break;
    case 'B':
      vc[4]++;
      break;
    }
  }
  if (np->type == XOR) {
    if (vc[2])
      return ('X');
    else if (!vc[3] && !vc[4])
      return ((vc[1] % 2) ? '1' : '0');
    else
      return (xor_val(vc));

  }
  cv = c ? '1' : '0';
  if (vc[c] || (vc[3] && vc[4]))
    rval = cv;
  else if (vc[c ^ 1] == np->fin)
    rval = cv ^ 1;
  else if (vc[2] > 0)
    rval = 'X';
  else if (vc[3])
    rval = 'D';
  else if (vc[4])
    rval = 'B';

  return (par ? inv(rval) : rval);
}

/*-----------------------------------------------------------------------
input: number of inputs that have 0, 1, X, D or B. 
this routine is called only when there are at least one D or B.
output: output value  
called by: simply and gimply
description:
-----------------------------------------------------------------------*/
char
xor_val(unsigned vc[])
{
  unsigned g_one, f_one;

  g_one = vc[1] + vc[3];
  f_one = vc[1] + vc[4];

  if (g_one == f_one)
    return ((g_one % 2) ? '1' : '0');
  else
    return ((g_one % 2) ? 'D' : 'B');
}

/*-----------------------------------------------------------------------
input: line value
output: the inverted line value
called by: gimply
description:
The routine inverts a line value using 5-valued logic simulation.
(D B 1 0 X) -> (B D 0 1 X)
-----------------------------------------------------------------------*/
char
inv(char lval)
{
  switch (lval) {
  case 'D':
    return ('B');
  case 'B':
    return ('D');
  case '1':
    return ('0');
  case '0':
    return ('1');
  default:
    return ('X');
  }
}

/*-----------------------------------------------------------------------
input: a pointer to a node and the implied value on the node
output: nothing
called by: podem
description:
This is a recursive routine to do all necessary implication.
It activates the fault if the target node is reached. The routine returns
if the line value is the same as the previous assignment, Otherwise, if a
boolean gate is reached, it calls "gimply" to do gate level implication
and then "imply" again with the output of the gate as the new arguments.
-----------------------------------------------------------------------*/
int
imply(VSTRUC np)
{
  unsigned int i;
  VSTRUC pp;

  if (np.node == Tgfault.node) {
    if (np.val == '1' && Tgfault.val == '0')
      np.val = 'D';
    else if (np.val == '0' && Tgfault.val == '1')
      np.val = 'B';
  }
  if (np.node->val != np.val) {
    np.node->val = np.val;
    for (i = 0; i < np.node->fout; i++) {
      pp.node = np.node->dnodes[i];
      pp.val = gimply(pp.node);
      if (pp.node->val != pp.val)
	imply(pp);
    }
  }
  return 0;
}

/*-----------------------------------------------------------------------
input: nothing
output: nothing
called by: main
description:
The routine drops all the faults detected by the current test vector
-----------------------------------------------------------------------*/

int
drop_fault(void)
{
  unsigned int i, j;
  NSTRUC *np;
  char val;
  fault_list *temp;

  for (i = 0; i < Npo; i++) {
    for (j = 0; j < Nnodes; j++)
      if (Bitset(Poutput[i]->flist, j)) {
	np = &Node[j];
	val = Node[j].val ^ 1;
	if ((temp = search_fault(np, val)) != NULL)
	  remove_fault(temp);
      }
  }
  return 0;
}

/*-----------------------------------------------------------------------
input: fault node pointer, type (stuck-at 0 or 1)
output: pointer to fault in the fault list
called by: drop_fault
description:
The routine searches the fault list for the given faults and returns
the pointer or null (if fault not present)
-----------------------------------------------------------------------*/

fault_list *
search_fault(NSTRUC *np, char val)
{
  fault_list *temp;

  temp = fault_set;

  while (temp != NULL) {
    if ((temp->node == np) && (temp->sa01 == val))
      return (temp);
    temp = temp->next;
  }
  return (temp);
}

/*-----------------------------------------------------------------------
input : a pointer to a target fault in list
output: nothing
called by: main, drop_fault
description:
This routine remove fault detected by fault simulation
------------------------------------------------------------------------*/
int
remove_fault(fault_list *new)
{

  if (new->next == NULL) {
    if (new->prev != NULL)
      new->prev->next = NULL;
    else
      fault_set = new->next;
  } else if (new->prev == NULL) {
    fault_set = new->next;
    if (new->next != NULL)
      new->next->prev = NULL;
  } else {
    new->prev->next = new->next;
    new->next->prev = new->prev;
  }
  free(new);
  return 0;
}

/*-----------------------------------------------------------------------
input : nothing
output: nothing
called by: main
description:
print information related to each line
these routines are used for debugging
------------------------------------------------------------------------*/
int
print_detected_faults(void)
{
  unsigned i, j;
  fault_list *temp;
  NSTRUC *np;

  j = 0;
  printf("Detected fault list\n");
  for (i = 0; i < Nnodes; i++) {
    if (j > 10) {
      printf("\n");
      j = 0;
    }

    np = &Node[i];
    if (!Bitset(det_sa0_fault, i)) {
      if ((temp = search_fault(np, '0')) != NULL) {
	printf("%d_%c ", i + 1, '0');
	j++;
      }
    }
    if (!Bitset(det_sa1_fault, i)) {
      if ((temp = search_fault(np, '1')) != NULL) {
	printf("%d_%c ", i + 1, '1');
	j++;
      }
    }
  }

  return 0;
}

/*-----------------------------------------------------------------------
input : nothing
output: nothing
called by: main
description:
print all undetected faults
------------------------------------------------------------------------*/
int
print_undetected_faults(void)
{
  unsigned i;

  /*
     printf("Undetected S-A-0 fault list\n");
   */
  for (i = 0; i < Nnodes; i++) {
    if (Bitset(det_sa0_fault, i))
      printf("%u 0\n", Node[i].num);
  }
  /*
     printf("Undetected S-A-1 fault list\n");
   */
  for (i = 0; i < Nnodes; i++) {
    if (Bitset(det_sa1_fault, i))
      printf("%u 1\n", Node[i].num);
  }
  return 0;
}

/*========================= End of program ============================*/
